""" MultiQC module to parse output from mirtop"""


import json
import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="mirtop",
            anchor="mirtop",
            href="https://github.com/miRTop/mirtop/",
            info="is a command line tool to annotate miRNAs and isomiRs and compute general statistics using the mirGFF3 format.",
            doi="10.5281/zenodo.45385",  # Zenodo won't load this page for me as I write this, but it's the listed DOI.
        )

        # Find and load any mirtop reports
        self.mirtop_data = dict()
        for f in self.find_log_files("mirtop"):
            self.parse_mirtop_report(f)
            self.add_data_source(f)

        # Filter out ignored samples (given with --ignore-samples option)
        self.mirtop_data = self.ignore_samples(self.mirtop_data)

        # Raise error if dict is empty
        if len(self.mirtop_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.mirtop_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.mirtop_data, "multiqc_mirtop")

        # Helper variables
        self.general_helptext = """
            The different isomiR types are:

            * `iso_3p`: a sequence with a 3' end difference because of trimming or templated tailing
            * `iso_5p`: a sequence with a 5' end difference because of trimming or templated tailing
            * `iso_add3p`: a sequence with non templated tailing in the 3' end
            * `iso_add5p`: a sequence with non templated tailing in the 5' end
            * `iso_snv`: a sequence with a single nucleotide variant

            The `ref_miRNA` label corresponds to the reference miRNA (canonical sequence).

        """
        self.isomir_cats = [
            "ref_miRNA",
            "iso_3p",
            "iso_5p",
            "iso_add3p",
            "iso_add5p",
            "iso_add",
            "iso_snv",
            "iso_snv_seed",
            "iso_snv_central_offset",
            "iso_snv_central",
            "iso_snv_central_supp",
            "iso_snp",
            "iso_snp_seed",
            "iso_snp_central_offset",
            "iso_snp_central",
            "iso_snp_central_supp",
        ]

        # Calculate aggregate iso_snp counts
        self.mirtop_data_snp_aggregate = self.aggregate_snps_in_samples()

        # Create summary table
        self.mirtop_stats_table()

        # Create detailed plots
        self.mirtop_read_count()
        self.mirtop_unique_read_count()
        self.mirtop_mean_read_count()

    def parse_mirtop_report(self, f):
        """Parse the mirtop log file."""

        content = json.loads(f["f"])
        version = content.get("meta", {}).get("version")

        for s_name in content.get("metrics", {}).keys():
            cleaned_s_name = self.clean_s_name(s_name, f)
            ## Check for sample name duplicates
            if cleaned_s_name in self.mirtop_data:
                log.debug(f"Duplicate sample name found! Overwriting: {cleaned_s_name}")
            parsed_data = content["metrics"][s_name]
            # Sum the isomiR and ref_miRNA counts if present.
            parsed_data["read_count"] = parsed_data.get("isomiR_sum", 0) + parsed_data.get("ref_miRNA_sum", 0)
            if parsed_data["read_count"] > 0:
                parsed_data["isomiR_perc"] = (parsed_data.get("isomiR_sum", 0) / parsed_data.get("read_count", 0)) * 100
            else:
                parsed_data["isomiR_perc"] = 0.0
            self.mirtop_data[cleaned_s_name] = parsed_data

            if version is not None:
                version = version.strip("v")
                self.add_software_version(version, cleaned_s_name)

    def aggregate_snps_in_samples(self):
        """Aggregate info for iso_snp isomiRs (for clarity). "Mean" section will be recomputed"""
        snv_aggr = {}  ## sub dict with all infos except for snps
        for sample in self.mirtop_data:
            snv_aggr[sample] = {
                key: self.mirtop_data[sample][key] for key in self.mirtop_data[sample] if "iso_snp" not in key
            }
            snv_aggr[sample]["iso_snv_sum"] = sum(
                [self.mirtop_data[sample][key] for key in self.mirtop_data[sample] if "iso_snp" in key and "sum" in key]
            )
            snv_aggr[sample]["iso_snv_count"] = sum(
                [
                    self.mirtop_data[sample][key]
                    for key in self.mirtop_data[sample]
                    if "iso_snp" in key and "count" in key
                ]
            )

            if snv_aggr[sample]["iso_snv_count"] > 0:
                snv_aggr[sample]["iso_snv_mean"] = snv_aggr[sample]["iso_snv_sum"] / snv_aggr[sample]["iso_snv_count"]
            else:
                snv_aggr[sample]["iso_snv_mean"] = 0
        return snv_aggr

    def mirtop_stats_table(self):
        """Take the parsed stats from the mirtop report and add them to the
        basic stats table at the top of the report"""

        headers = {
            "ref_miRNA_sum": {
                "title": f"{config.read_count_prefix} Ref miRNA reads",
                "description": f"Read counts summed over all reference miRNAs ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "PuBu",
            },
            "isomiR_perc": {
                "title": "IsomiR %",
                "description": "% of total read counts corresponding to isomiRs",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "YlOrRd",
            },
            "isomiR_sum": {
                "title": f"{config.read_count_prefix} IsomiR reads",
                "description": f"Read counts summed over all isomiRs ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "Oranges",
            },
            "read_count": {
                "title": f"{config.read_count_prefix} Reads",
                "description": "Total read counts - both isomiRs and reference miRNA ({})".format(
                    config.read_count_desc
                ),
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "BuGn",
            },
        }

        self.general_stats_addcols(self.mirtop_data, headers)

    def filter_plot_data(self, plot_key):
        """Helper function to return the subset of data that has been requested for this plot"""
        plot_data = {}
        for s_name in self.mirtop_data_snp_aggregate:
            plot_data[s_name] = {}
            for key, data in self.mirtop_data_snp_aggregate[s_name].items():
                if plot_key in key and "isomiR" not in key and "read" not in key:
                    plot_data[s_name][key] = data
        return plot_data

    def get_plot_cats(self, plot_type):
        """Return the plot categories for the given plot"""
        cats_section = dict()
        for base_key in self.isomir_cats:
            cat_key = f"{base_key}_{plot_type}"
            cats_section[cat_key] = {"name": base_key}
        return cats_section

    def mirtop_read_count(self):
        """Generate barplot for the read count plot"""
        p_config = {"id": "mirtop_read_count_plot", "title": "mirtop: IsomiR read counts", "ylab": "Read counts"}
        self.add_section(
            name="IsomiR read counts",
            description="""
            Total counts of reads aligned for each isomiR type, over all detected miRNAs.
            """,
            helptext="""
            The total counts of reads detected as reference miRNA sequences is also shown.

            Since a read can belong to 2 (or more) different isomiRs types (e.g `iso_3p` and `iso_5p`),
            the cumulative read counts shown in this plot for a sample can be higher than its total
            read count shown in the general statistics.

            For each sample, the mean counts of each type of isomiRs over all detected
            miRNAs is displayed in a different color.
            """
            + self.general_helptext,
            plot=bargraph.plot(self.filter_plot_data("sum"), self.get_plot_cats("sum"), p_config),
        )

    def mirtop_unique_read_count(self):
        """Generate the section for the Unique Read Count plot"""
        p_config = {
            "id": "mirtop_unique_read_count_plot",
            "title": "mirtop: IsomiR unique read counts",
            "ylab": "Unique sequences",
        }
        self.add_section(
            name="IsomiR unique read counts",
            anchor="mirtop_unique_read_count",
            description="""
            The number of distinct sequences detected for each isomiR type, over all miRNAs.
            """,
            helptext="""
            The number of reference miRNA sequences detected is also shown.

            For each sample, the number of miRNAs with each type of isomiRs, is displayed in a different color.
            """
            + self.general_helptext,
            plot=bargraph.plot(self.filter_plot_data("count"), self.get_plot_cats("count"), p_config),
        )

    def mirtop_mean_read_count(self):
        """Generate the section for the Mean Read Count plot"""
        p_config = {"id": "mirtop_mean_read_count_plot", "title": "mirtop: Mean isomiR read counts", "ylab": "Means"}
        self.add_section(
            name="Mean isomiR read counts",
            anchor="mirtop_mean_read_count",
            description="""
            Mean counts for each isomiR type, over all detected miRNAs.
            """,
            helptext="""
            The mean counts of reads detected as reference miRNA sequences is also shown.

            For each sample, the mean counts of each type of isomiRs over all detected miRNAs is displayed in a different color.
            """
            + self.general_helptext,
            plot=bargraph.plot(self.filter_plot_data("mean"), self.get_plot_cats("mean"), p_config),
        )



# -*- coding: utf-8 -*-

""" MultiQC module to parse output from phantompeakqualtools """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="phantompeakqualtools",
            anchor="phantompeakqualtools",
            href="https://www.encodeproject.org/software/phantompeakqualtools",
            info="computes informative enrichment and quality measures for ChIP-seq/DNase-seq/FAIRE-seq/MNase-seq data.",
            doi=["10.1101/gr.136184.111", "10.1038/nbt.1508"],
        )

        # Parse logs
        self.phantompeakqualtools_data = dict()
        for f in self.find_log_files("phantompeakqualtools/out", filehandles=False):
            self.parse_phantompeakqualtools(f)

        # Filter to strip out ignored sample names
        self.phantompeakqualtools_data = self.ignore_samples(self.phantompeakqualtools_data)

        # Warning when no files are found
        if len(self.phantompeakqualtools_data) == 0:
            raise ModuleNoSamplesFound

        # Log
        log.info(f"Found {len(self.phantompeakqualtools_data)} logs")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None, f["s_name"])

        # Write parsed data to a file
        self.write_data_file(self.phantompeakqualtools_data, "multiqc_phantompeakqualtools")

        # Report section
        self.phantompeakqualtools_general_stats()

    # Parse spp.out file from phantompeakqualtools
    def parse_phantompeakqualtools(self, f):
        parsed_data = {}
        lines = f["f"].splitlines()
        for line in lines:
            s = line.split("\t")
            parsed_data["Estimated_Fragment_Length_bp"] = int(s[2].split(",")[0])
            parsed_data["NSC"] = float(s[8])
            parsed_data["RSC"] = float(s[9])
        if len(parsed_data) > 0:
            if f["s_name"] in self.phantompeakqualtools_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f)
            self.phantompeakqualtools_data[f["s_name"]] = parsed_data

    # Report fragment length, NSC and RSC in general stat table
    def phantompeakqualtools_general_stats(self):
        """Add columns to General Statistics table"""
        headers = {
            "Estimated_Fragment_Length_bp": {
                "title": "Frag Length",
                "description": "Estimated fragment length (bp)",
                "min": 0,
                "format": "{:,.0f}",
            },
            "NSC": {
                "title": "NSC",
                "description": "Normalized strand cross-correlation",
                "max": 10,
                "min": 0,
                "format": "{:,.2f}",
                "scale": "RdYlGn-rev",
            },
            "RSC": {
                "title": "RSC",
                "description": "Relative strand cross-correlation",
                "max": 10,
                "min": 0,
                "format": "{:,.2f}",
                "scale": "RdYlBu-rev",
            },
        }
        self.general_stats_addcols(self.phantompeakqualtools_data, headers)



#!/usr/bin/env python

""" MultiQC code to export data to MegaQC / flat JSON files """


import gzip
import io
import json

import requests

from . import config
from .util_functions import MQCJSONEncoder

log = config.logger


def multiqc_api_post(exported_data):
    headers = {"Content-Type": "application/json", "content-encoding": "gzip"}
    if config.megaqc_access_token is not None:
        headers["access_token"] = config.megaqc_access_token
    post_data = json.dumps({"data": exported_data}, cls=MQCJSONEncoder, ensure_ascii=False, indent=2)
    post_data = post_data.encode("utf-8", "ignore")

    # Gzip the JSON for massively decreased filesize
    sio_obj = io.BytesIO()
    gzfh = gzip.GzipFile(fileobj=sio_obj, mode="w")
    gzfh.write(post_data)
    gzfh.close()
    request_body = sio_obj.getvalue()

    log.debug("Sending data to MegaQC")
    log.debug(f"MegaQC URL: {config.megaqc_url}")
    try:
        r = requests.post(config.megaqc_url, headers=headers, data=request_body, timeout=config.megaqc_timeout)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout) as e:
        log.error(f"Timed out when sending data: {e}")
    except requests.exceptions.ConnectionError:
        log.error(f"Couldn't connect to MegaQC URL {config.megaqc_url}")
    except Exception as e:
        log.error(f"Error sending data: {e}")
    else:
        try:
            api_r = json.loads(r.text)
        except Exception:
            log.error(f"Error: JSON response could not be parsed (status code: {r.status_code})")
            return None
        if r.status_code == 200:
            if api_r["success"]:
                log.info(f"{api_r['message']}")
            else:
                log.error(f"Error - {api_r['message']}")
        else:
            if r.status_code == 403:
                if config.megaqc_access_token is not None:
                    log.error("Error 403: Authentication error, megaqc_access_token not recognised")
                else:
                    log.error("Error 403: Authentication error, megaqc_access_token is required")
            else:
                log.debug(f"MegaQC API status code was {r.status_code}")
                log.error(f"Error - {api_r.get('message', 'Unknown problem')}")



""" MultiQC module to parse output from MACS2 """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="MACS2",
            anchor="macs",
            href="https://macs3-project.github.io/MACS/",
            info="identifies transcription factor binding sites in ChIP-seq data.",
            doi=["10.1101/496521", "10.1186/gb-2008-9-9-r137"],
        )

        # Parse logs
        self.macs_data = dict()
        for f in self.find_log_files("macs2", filehandles=True):
            self.parse_macs(f)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.macs_data = self.ignore_samples(self.macs_data)

        if len(self.macs_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.macs_data)} logs")
        self.write_data_file(self.macs_data, "multiqc_macs")

        self.macs_general_stats()
        self.macs_filtered_reads_plot()

    def parse_macs(self, f):
        regexes = {
            "name": r"# name = (.+)$",
            "fragment_size": r"# (?:fragment|tag) size is determined as (\d+) bps",
            "treatment_fragments_total": r"# total (?:fragments|tags) in treatment: (\d+)",
            "treatment_fragments_after_filtering": r"# (?:fragments|tags) after filtering in treatment: (\d+)",
            "treatment_max_duplicates": r"# maximum duplicate (?:fragments|tags at the same position) in treatment = (\d+)",
            "treatment_redundant_rate": r"# Redundant rate in treatment: ([\d\.]+)",
            "control_fragments_total": r"# total (?:fragments|tags) in control: (\d+)",
            "control_fragments_after_filtering": r"# (?:fragments|tags) after filtering in control: (\d+)",
            "control_max_duplicates": r"# maximum duplicate (?:fragments|tags at the same position) in control = (\d+)",
            "control_redundant_rate": r"# Redundant rate in control: ([\d\.]+)",
            "d": r"# d = (\d+)",
            "version": r"# This file is generated by MACS version ([\d\.]+)",
        }
        s_name = f["s_name"]
        parsed_data = {"peak_count": 0}
        version = None
        for line in f["f"]:
            line = line.strip()
            if line.startswith("#"):
                for k, r in regexes.items():
                    match = re.search(r, line)
                    if match:
                        if k == "name":
                            s_name = self.clean_s_name(match.group(1).strip(), f)
                        elif k == "version":
                            version = match.group(1)
                        else:
                            parsed_data[k] = float(match.group(1).strip())
            elif len(line) > 0 and "start" not in line:
                parsed_data["peak_count"] += 1

        if len(parsed_data) > 0:
            if s_name in self.macs_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.macs_data[s_name] = parsed_data
            if version is not None:
                self.add_software_version(version, s_name)

    def macs_general_stats(self):
        """Add columns to General Statistics table"""
        headers = {
            "d": {"title": "Fragment Length", "min": 0, "format": "{:,.0f}"},
            "treatment_redundant_rate": {
                "title": "Treatment Redundancy",
                "description": "Redundant rate in treatment",
                "max": 1,
                "min": 0,
                "format": "{:,.2f}",
                "scale": "RdYlBu-rev",
            },
            "control_redundant_rate": {
                "title": "Control Redundancy",
                "description": "Redundant rate in control",
                "max": 1,
                "min": 0,
                "format": "{:,.2f}",
                "scale": "RdYlBu-rev",
            },
            "peak_count": {
                "title": "Number of Peaks",
                "description": "Total number of peaks",
                "min": 0,
                "format": "{:,.0f}",
            },
        }
        self.general_stats_addcols(self.macs_data, headers)

    def macs_filtered_reads_plot(self):
        """Plot of filtered reads for control and treatment samples"""
        data = dict()
        req_cats = [
            "control_fragments_total",
            "control_fragments_after_filtering",
            "treatment_fragments_total",
            "treatment_fragments_after_filtering",
        ]
        for s_name, d in self.macs_data.items():
            if all([c in d for c in req_cats]):
                data[f"{s_name}: Control"] = dict()
                data[f"{s_name}: Treatment"] = dict()
                data[f"{s_name}: Control"]["fragments_filtered"] = (
                    d["control_fragments_total"] - d["control_fragments_after_filtering"]
                )
                data[f"{s_name}: Control"]["fragments_not_filtered"] = d["control_fragments_after_filtering"]
                data[f"{s_name}: Treatment"]["fragments_filtered"] = (
                    d["treatment_fragments_total"] - d["treatment_fragments_after_filtering"]
                )
                data[f"{s_name}: Treatment"]["fragments_not_filtered"] = d["treatment_fragments_after_filtering"]

        # Check that we have something to plot
        if len(data) == 0:
            return

        # Specify the order of the different possible categories
        keys = {
            "fragments_not_filtered": {"color": "#437BB1", "name": "Remaining fragments"},
            "fragments_filtered": {"color": "#B1084C", "name": "Filtered fragments"},
        }

        # Config for the plot
        pconfig = {
            "id": "macs2_filtered",
            "title": "MACS2: Filtered Fragments",
            "ylab": "# Fragments",
            "cpswitch_counts_label": "Number of Fragments",
            "hide_zero_cats": False,
        }

        self.add_section(plot=bargraph.plot(data, keys, pconfig))



import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class DragenRnaQuantMetrics(BaseMultiqcModule):
    def add_rna_metrics(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/rna_quant_metrics"):
            data = parse_time_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="rna_quant_metrics")
            data_by_sample[s_name] = data

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)

        if not data_by_sample:
            return set()

        # Only plot data, don't want to write this to a file
        # (can do so with --export-plots already)
        # self.write_data_file(data_by_sample, "dragen_quant_metrics")

        metric_names = {
            "Transcript fragments",
            "Strand mismatched fragments",
            "Ambiguous strand fragments",
            "Unknown transcript fragments",
            "Intron fragments",
            "Intergenic fragments",
        }

        self.add_section(
            name="RNA Quantification Metrics",
            anchor="dragen-rna-quant-metrics",
            description="RNA quantification metrics for DRAGEN.",
            plot=bargraph.plot(
                [
                    {
                        sample: {metric: stat for metric, stat in data.items() if metric in metric_names}
                        for sample, data in data_by_sample.items()
                    }
                ],
                pconfig={
                    "id": "dragen_rna_quant_metrics",
                    "title": "Dragen: RNA Quant Metrics",
                    "ylab": "Fragments",
                    "cpswitch_counts_label": "Fragments",
                    "data_labels": [
                        {
                            "name": "Transcript / Intronic / Intergenic Fragments",
                            "ylab": "Fragments",
                            "cpswitch_counts_label": "Fragments",
                        },
                        {
                            "name": "Fragment Orientations",
                            "ylab": "Fragments",
                            "cpswitch_counts_label": "Fragments",
                        },
                    ],
                },
            ),
        )

        return data_by_sample.keys()


def parse_time_metrics_file(f):
    """
    sample.time_metrics.csv

    RUN TIME,,Time loading reference,00:01:31.289,91.29
    RUN TIME,,Time aligning reads,00:00:25.190,25.19
    RUN TIME,,Time duplicate marking,00:00:01.817,1.82
    RUN TIME,,Time sorting and marking duplicates,00:00:07.368,7.37
    RUN TIME,,Time DRAGStr calibration,00:00:07.069,7.07
    """
    data = {}
    for line in f["f"].splitlines():
        tokens = line.split(",")
        if len(tokens) == 4:
            analysis, _, metric, stat = tokens
            percentage = None
        elif len(tokens) == 5:
            analysis, _, metric, stat, percentage = tokens
        else:
            raise ValueError(f"Unexpected number of tokens in line {line}")

        try:
            stat = float(stat)
        except ValueError:
            pass
        data[metric] = stat

    return data



""" MultiQC submodule to parse output from Picard HsMetrics """

import logging
from collections import defaultdict

from multiqc import config
from multiqc.modules.picard import util
from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)

FIELD_DESCRIPTIONS = {
    "AT_DROPOUT": "A measure of how undercovered <= 50% GC regions are relative to the mean.",
    "BAIT_DESIGN_EFFICIENCY": "Target territory / bait territory. 1 == perfectly efficient, 0.5 = half of baited bases are not target.",
    "BAIT_SET": "The name of the bait set used in the hybrid selection.",
    "BAIT_TERRITORY": "The number of bases which have one or more baits on top of them.",
    "FOLD_80_BASE_PENALTY": 'The fold over-coverage necessary to raise 80% of bases in "non-zero-cvg" targets to the mean coverage level in those targets.',
    "FOLD_ENRICHMENT": "The fold by which the baited region has been amplified above genomic background.",
    "GC_DROPOUT": "A measure of how undercovered >= 50% GC regions are relative to the mean.",
    "GENOME_SIZE": "The number of bases in the reference genome used for alignment.",
    "HET_SNP_Q": "The Phred Scaled Q Score of the theoretical HET SNP sensitivity.",
    "HET_SNP_SENSITIVITY": "The theoretical HET SNP sensitivity.",
    "HS_LIBRARY_SIZE": "The estimated number of unique molecules in the selected part of the library.",
    "HS_PENALTY_100X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 100X. This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 100X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 100 * HS_PENALTY_100X.",
    "HS_PENALTY_10X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 10X. This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 10X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 10 * HS_PENALTY_10X.",
    "HS_PENALTY_20X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 20X. This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 20X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 20 * HS_PENALTY_20X.",
    "HS_PENALTY_30X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 30X. This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 30X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 30 * HS_PENALTY_30X.",
    "HS_PENALTY_40X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 40X.  This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 40X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 40 * HS_PENALTY_40X.",
    "HS_PENALTY_50X": "The 'hybrid selection penalty' incurred to get 80% of target bases to 50X.  This metric should be interpreted as: if I have a design with 10 megabases of target, and want to get 50X coverage I need to sequence until PF_ALIGNED_BASES = 10^7 * 50 * HS_PENALTY_50X.",
    "MAX_TARGET_COVERAGE": "The maximum coverage of reads that mapped to target regions of an experiment.",
    "MEAN_BAIT_COVERAGE": "The mean coverage of all baits in the experiment.",
    "MEAN_TARGET_COVERAGE": "The mean coverage of targets.",
    "MEDIAN_TARGET_COVERAGE": "The median coverage of targets.",
    "NEAR_BAIT_BASES": "The number of PF aligned bases that mapped to within a fixed interval of a baited region, but not on a baited region.",
    "OFF_BAIT_BASES": "The number of PF aligned bases that mapped to neither on or near a bait.",
    "OLD_80_BASE_PENALTY": 'The fold over-coverage necessary to raise 80% of bases in "non-zero-cvg" targets to the mean coverage level in those targets.',
    "ON_BAIT_BASES": "The number of PF aligned bases that mapped to a baited region of the genome.",
    "ON_BAIT_VS_SELECTED": "The percentage of on+near bait bases that are on as opposed to near.",
    "ON_TARGET_BASES": "The number of PF aligned bases that mapped to a targeted region of the genome.",
    "PCT_EXC_BASEQ": "The fraction of aligned bases that were filtered out because they were of low base quality.",
    "PCT_EXC_DUPE": "The fraction of aligned bases that were filtered out because they were in reads marked as duplicates.",
    "PCT_EXC_MAPQ": "The fraction of aligned bases that were filtered out because they were in reads with low mapping quality.",
    "PCT_EXC_OFF_TARGET": "The fraction of aligned bases that were filtered out because they did not align over a target base.",
    "PCT_EXC_OVERLAP": "The fraction of aligned bases that were filtered out because they were the second observation from an insert with overlapping reads.",
    "PCT_OFF_BAIT": "The percentage of aligned PF bases that mapped neither on or near a bait.",
    "PCT_PF_READS": "PF reads / total reads. The percent of reads passing filter.",
    "PCT_PF_UQ_READS_ALIGNED": "PF Reads Aligned / PF Reads.",
    "PCT_PF_UQ_READS": "PF Unique Reads / Total Reads.",
    "PCT_SELECTED_BASES": "On+Near Bait Bases / PF Bases Aligned.",
    "PCT_TARGET_BASES_100X": "The fraction of all target bases achieving 100X or greater coverage.",
    "PCT_TARGET_BASES_10X": "The fraction of all target bases achieving 10X or greater coverage.",
    "PCT_TARGET_BASES_1X": "The fraction of all target bases achieving 1X or greater coverage.",
    "PCT_TARGET_BASES_20X": "The fraction of all target bases achieving 20X or greater coverage.",
    "PCT_TARGET_BASES_2X": "The fraction of all target bases achieving 2X or greater coverage.",
    "PCT_TARGET_BASES_30X": "The fraction of all target bases achieving 30X or greater coverage.",
    "PCT_TARGET_BASES_40X": "The fraction of all target bases achieving 40X or greater coverage.",
    "PCT_TARGET_BASES_50X": "The fraction of all target bases achieving 50X or greater coverage.",
    "PCT_USABLE_BASES_ON_BAIT": "The number of aligned, de-duped, on-bait bases out of the PF bases available.",
    "PCT_USABLE_BASES_ON_TARGET": "The number of aligned, de-duped, on-target bases out of the PF bases available.",
    "PF_BASES_ALIGNED": "The number of PF unique bases that are aligned with mapping score > 0 to the reference genome.",
    "PF_READS": "The number of reads that pass the vendor's filter.",
    "PF_UNIQUE_READS": "The number of PF reads that are not marked as duplicates.",
    "PF_UQ_BASES_ALIGNED": "The number of bases in the PF aligned reads that are mapped to a reference base. Accounts for clipping and gaps.",
    "PF_UQ_READS_ALIGNED": "The number of PF unique reads that are aligned with mapping score > 0 to the reference genome.",
    "TARGET_TERRITORY": "The unique number of target bases in the experiment where target is usually exons etc.",
    "TOTAL_READS": "The total number of reads in the SAM or BAM file examine.",
    "ZERO_CVG_TARGETS_PCT": "The fraction of targets that did not reach coverage=1 over any base.",
}


def parse_reports(module):
    """Find Picard HsMetrics reports and parse their data"""

    data_by_bait_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/hsmetrics", filehandles=True):
        s_name = f["s_name"]
        keys = None
        commadecimal = None

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectHsMetrics",
                sentieon_algo="HsMetricAlgo",
            )
            if maybe_s_name:
                s_name = maybe_s_name
                keys = None

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="HsMetrics", sentieon_algo="HsMetricAlgo"):
                keys = f["f"].readline().strip("\n").split("\t")
                if s_name in data_by_bait_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                data_by_bait_by_sample[s_name] = dict()

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys):
                    keys = None
                    continue

                bait = "NA"
                if keys[0] == "BAIT_SET":
                    bait = vals[0]
                data_by_bait_by_sample[s_name][bait] = dict()
                # Check that we're not using commas for decimal places
                if commadecimal is None:
                    commadecimal = False
                    for i, k in enumerate(keys):
                        if "PCT" in k or "BAIT" in k or "MEAN" in k:
                            if "," in vals[i]:
                                commadecimal = True
                                break
                for i, k in enumerate(keys):
                    try:
                        if commadecimal:
                            vals[i] = vals[i].replace(".", "")
                            vals[i] = vals[i].replace(",", ".")
                        data_by_bait_by_sample[s_name][bait][k] = float(vals[i])
                    except ValueError:
                        data_by_bait_by_sample[s_name][bait][k] = vals[i]

    # Remove empty dictionaries
    for s_name in data_by_bait_by_sample:
        for bait in data_by_bait_by_sample[s_name]:
            if len(data_by_bait_by_sample[s_name][bait]) == 0:
                data_by_bait_by_sample[s_name].pop(bait, None)
        if len(data_by_bait_by_sample[s_name]) == 0:
            data_by_bait_by_sample.pop(s_name, None)

    data_by_sample = dict()
    # Manipulate sample names if multiple baits found
    for s_name in data_by_bait_by_sample:
        for bait in data_by_bait_by_sample[s_name]:
            s_bait_name = s_name
            # If there are multiple baits, append the bait name to the sample name
            if len(data_by_bait_by_sample[s_name]) > 1:
                s_bait_name = f"{s_name}: {bait}"
            if s_bait_name in data_by_sample:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_bait_name}")
            data_by_sample[s_bait_name] = data_by_bait_by_sample[s_name][bait]
            module.add_data_source(f, s_bait_name, section="HsMetrics")

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_HsMetrics")

    # Swap question marks with -1
    for s_name in data_by_sample:
        if data_by_sample[s_name]["FOLD_ENRICHMENT"] == "?":
            data_by_sample[s_name]["FOLD_ENRICHMENT"] = -1

    # Add to general stats table
    _general_stats_table(module, data_by_sample)

    # Add report section
    module.add_section(
        name="HSMetrics",
        anchor=f"{module.anchor}_hsmetrics",
        plot=table.plot(
            data_by_sample,
            _get_table_headers(),
            {
                "id": f"{module.anchor}_hsmetrics_table",
                "namespace": "HsMetrics",
                "scale": "RdYlGn",
                "min": 0,
                "title": "Picard HsMetrics",
            },
        ),
    )
    tbases = _add_target_bases(module, data_by_sample)
    module.add_section(
        name=tbases["name"],
        anchor=tbases["anchor"],
        description=tbases["description"],
        plot=tbases["plot"],
    )
    hs_pen_plot = hs_penalty_plot(module, data_by_sample)
    if hs_pen_plot is not None:
        module.add_section(
            name="HS Penalty",
            anchor=f"{module.anchor}_hsmetrics_hs_penalty",
            description='The "hybrid selection penalty" incurred to get 80% of target bases to a given coverage.',
            helptext="""
                Can be used with the following formula:

                ```
                required_aligned_bases = bait_size_bp * desired_coverage * hs_penalty
                ```
            """,
            plot=hs_pen_plot,
        )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)


def _general_stats_table(module, data):
    """
    Generate table header configs for the General Stats table,
    add config and data to the base module.
    """
    # Look for a user config of which table columns we should use
    picard_config = getattr(config, "picard_config", {})
    genstats_table_cols = picard_config.get("HsMetrics_genstats_table_cols", [])
    genstats_table_cols_hidden = picard_config.get("HsMetrics_genstats_table_cols_hidden", [])

    headers = {}
    # Custom general stats columns
    if len(genstats_table_cols) or len(genstats_table_cols_hidden):
        for k, v in _generate_table_header_config(genstats_table_cols, genstats_table_cols_hidden).items():
            headers[k] = v

    # Default General Stats headers
    else:
        headers["FOLD_ENRICHMENT"] = {
            "title": "Fold Enrichment",
            "min": 0,
            "format": "{:,.0f}",
            "scale": "Blues",
            "suffix": " X",
        }
        headers["MEDIAN_TARGET_COVERAGE"] = {
            "title": "Median Target Coverage",
            "description": "The median coverage of reads that mapped to target regions of an experiment.",
            "min": 0,
            "suffix": "X",
            "scale": "GnBu",
        }
        try:
            covs = picard_config["general_stats_target_coverage"]
            assert isinstance(covs, list)
            assert len(covs) > 0
            covs = [str(i) for i in covs]
            log.debug(f"Custom picad coverage thresholds: {', '.join([i for i in covs])}")
        except (KeyError, AttributeError, TypeError, AssertionError):
            covs = ["30"]
        for c in covs:
            headers[f"PCT_TARGET_BASES_{c}X"] = {
                "id": f"{module.anchor}_target_bases_{c}X",
                "title": f"Target Bases ≥ {c}X",
                "description": f"Percent of target bases with coverage ≥ {c}X",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "format": "{:,.0f}",
                "scale": "RdYlGn",
                "modify": lambda x: util.multiply_hundred(x),
            }
    module.general_stats_addcols(data, headers, namespace="HsMetrics")


def _get_table_headers():
    # Look for a user config of which table columns we should use
    picard_config = getattr(config, "picard_config", {})
    HsMetrics_table_cols = picard_config.get("HsMetrics_table_cols")
    HsMetrics_table_cols_hidden = picard_config.get("HsMetrics_table_cols_hidden")

    # Default table columns
    if not HsMetrics_table_cols:
        HsMetrics_table_cols = [
            "AT_DROPOUT",
            "BAIT_DESIGN_EFFICIENCY",
            "BAIT_TERRITORY",
            "FOLD_80_BASE_PENALTY",
            "FOLD_ENRICHMENT",
            "GC_DROPOUT",
            "HET_SNP_Q",
            "HET_SNP_SENSITIVITY",
            "NEAR_BAIT_BASES",
            "OFF_BAIT_BASES",
            "ON_BAIT_BASES",
            "ON_TARGET_BASES",
            "PCT_USABLE_BASES_ON_BAIT",
            "PCT_USABLE_BASES_ON_TARGET",
            "PF_BASES_ALIGNED",
            "PF_READS",
            "PCT_SELECTED_BASES",
            "PF_UNIQUE_READS",
            "PF_UQ_BASES_ALIGNED",
            "PF_UQ_READS_ALIGNED",
            "TOTAL_READS",
            "MAX_TARGET_COVERAGE",
            "MEAN_BAIT_COVERAGE",
            "MEAN_TARGET_COVERAGE",
            "MEDIAN_TARGET_COVERAGE",
            "ON_BAIT_VS_SELECTED",
            "TARGET_TERRITORY",
            "ZERO_CVG_TARGETS_PCT",
        ]
    if not HsMetrics_table_cols_hidden:
        HsMetrics_table_cols_hidden = [
            "BAIT_TERRITORY",
            "TOTAL_READS",
            "TARGET_TERRITORY",
            "AT_DROPOUT",
            "GC_DROPOUT",
        ]

    return _generate_table_header_config(HsMetrics_table_cols, HsMetrics_table_cols_hidden)


def _generate_table_header_config(table_cols, hidden_table_cols):
    """
    Automatically generate some nice table header configs based on what we know about
    the different types of Picard data fields.
    """
    title_cleanup = [
        ("CVG", "coverage"),
        ("UQ", "unique"),
        ("ON_", "On-"),
        ("OFF_", "Off-"),
        ("NEAR_", "Near-"),
        ("_", " "),
        ("PCT", ""),
    ]

    # Warn if we see anything unexpected
    for c in table_cols + hidden_table_cols:
        if c not in FIELD_DESCRIPTIONS and c[:17] != "PCT_TARGET_BASES_":
            log.error(f"Field '{c}' not found in expected Picard fields. Please check your config.")

    headers = dict()
    for h in table_cols + hidden_table_cols:
        # Set up the configuration for each column
        if h not in headers:
            # Generate a nice string for the column title
            h_title = h
            for s, r in title_cleanup:
                h_title = h_title.replace(s, r)

            headers[h] = {
                "title": h_title.strip().lower().capitalize(),
                "description": FIELD_DESCRIPTIONS[h] if h in FIELD_DESCRIPTIONS else None,
            }
            if h.find("PCT") > -1:
                headers[h]["title"] = headers[h]["title"]
                headers[h]["modify"] = lambda x: x * 100.0
                headers[h]["max"] = 100
                headers[h]["suffix"] = "%"

            elif h.find("READS") > -1:
                headers[h]["title"] = f"{config.read_count_prefix} {headers[h]['title']}"
                headers[h]["modify"] = lambda x: x * config.read_count_multiplier
                headers[h]["shared_key"] = "read_count"

            elif h.find("BASES") > -1:
                headers[h]["title"] = f"{config.base_count_prefix} {headers[h]['title']}"
                headers[h]["modify"] = lambda x: x * config.base_count_multiplier
                headers[h]["shared_key"] = "base_count"

            # Manual capitilisation for some strings
            headers[h]["title"] = headers[h]["title"].replace("Pf", "PF").replace("snp", "SNP")

            if h in hidden_table_cols:
                headers[h]["hidden"] = True

    return headers


def _add_target_bases(self, data):
    data_clean = defaultdict(dict)
    max_non_zero_cov = 0
    for s in data:
        for h in data[s]:
            if h.startswith("PCT_TARGET"):
                cov = int(h.replace("PCT_TARGET_BASES_", "")[:-1])
                bases_pct = data[s][h]
                data_clean[s][cov] = bases_pct * 100.0
                if bases_pct > 0 and cov > max_non_zero_cov:
                    max_non_zero_cov = cov

    pconfig = {
        "id": f"{self.anchor}_percentage_target_bases",
        "title": f"{self.name}: Percentage of target bases",
        "xlab": "Fold Coverage",
        "ylab": "Pct of bases",
        "ymax": 100,
        "ymin": 0,
        "xmin": 0,
        "xmax": max_non_zero_cov,
        "tt_label": "<b>{point.x}X</b>: {point.y:.2f}%",
    }
    return {
        "name": "Target Region Coverage",
        "anchor": f"{self.anchor}_hsmetrics_target_bases",
        "description": "The percentage of all target bases with at least <code>x</code> fold coverage.",
        "plot": linegraph.plot(data_clean, pconfig),
    }


def hs_penalty_plot(self, data):
    data_clean = defaultdict(dict)
    any_non_zero = False
    for s in data:
        for h in data[s]:
            if h.startswith("HS_PENALTY"):
                data_clean[s][int(h.lstrip("HS_PENALTY_").rstrip("X"))] = data[s][h]
                if data[s][h] > 0:
                    any_non_zero = True

    pconfig = {
        "id": f"{self.anchor}_hybrid_selection_penalty",
        "title": f"{self.name}: Hybrid Selection Penalty",
        "xlab": "Fold Coverage",
        "ylab": "Penalty",
        "ymin": 0,
        "xmin": 0,
        "xDecimals": False,
        "tt_label": "<b>{point.x}X</b>: {point.y:.2f}%",
    }

    if any_non_zero:
        return linegraph.plot(data_clean, pconfig)



""" MultiQC module to parse output from DamageProfiler """


import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        """
        damageprofiler module class
        """

        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="DamageProfiler",
            anchor="damageprofiler",
            href="https://github.com/Integrative-Transcriptomics/DamageProfiler",
            info="a tool to determine damage patterns on ancient DNA.",
            doi="10.1093/bioinformatics/btab190",
        )

        # Init empty dictionaries
        self.threepGtoAfreq_data = dict()
        self.fivepCtoTfreq_data = dict()
        self.lgdist_fw_data = dict()
        self.lgdist_rv_data = dict()
        self.summary_metrics_data = dict()

        # Find and load JSON file
        for f in self.find_log_files("damageprofiler", filehandles=True):
            self.parseJSON(f)

        # Filter to strip out ignored sample names
        self.threepGtoAfreq_data = self.ignore_samples(self.threepGtoAfreq_data)
        self.fivepCtoTfreq_data = self.ignore_samples(self.fivepCtoTfreq_data)
        self.lgdist_fw_data = self.ignore_samples(self.lgdist_fw_data)
        self.lgdist_rv_data = self.ignore_samples(self.lgdist_rv_data)
        self.summary_metrics_data = self.ignore_samples(self.summary_metrics_data)

        if len(self.summary_metrics_data) == 0:
            raise ModuleNoSamplesFound

        # Write parsed report data to a file
        self.write_data_file(self.summary_metrics_data, "multiqc_damageprofiler_metrics")

        # Basic Stats Table, use generic function to add data to general table
        self.dmgprof_misinc_stats(self.threepGtoAfreq_data, "3 Prime", "G>A")
        self.dmgprof_misinc_stats(self.fivepCtoTfreq_data, "5 Prime", "C>T")
        self.add_summary_metrics(self.summary_metrics_data)

        # Add plots
        if len(self.threepGtoAfreq_data) > 0:
            self.add_section(
                name="3P misincorporation plot",
                description="3' misincorporation plot for G>A substitutions",
                helptext="""
                This plot shows the frequency of G>A substitutions at the 3' read ends. Typically, one would observe high substitution percentages for ancient DNA, whereas modern DNA does not show these in higher extents.
                """,
                plot=self.threeprime_plot(),
            )
        if len(self.fivepCtoTfreq_data) > 0:
            self.add_section(
                name="5P misincorporation plot",
                description="5' misincorporation plot for C>T substitutions",
                helptext="""
                This plot shows the frequency of C>T substitutions at the 5' read ends. Typically, one would observe high substitution percentages for ancient DNA, whereas modern DNA does not show these in higher extents.
                """,
                plot=self.fiveprime_plot(),
            )
        if len(self.lgdist_fw_data) > 0 and len(self.lgdist_rv_data) > 0:
            self.add_section(
                name="Forward read length distribution",
                description="Read length distribution for forward strand (+) reads.",
                helptext="""
                This plot shows the read length distribution of the forward reads in the investigated sample. Reads below lengths of 30bp are typically filtered, so the plot doesn't show these in many cases. A shifted distribution of read lengths towards smaller read lengths (e.g around 30-50bp) is also an indicator of ancient DNA.
                """,
                plot=self.lgdistplot(self.lgdist_fw_data, "Forward"),
            )
            self.add_section(
                name="Reverse read length distribution",
                description="Read length distribution for reverse strand (-) reads.",
                helptext="""
                This plot shows the read length distribution of the reverse reads in the investigated sample. Reads below lengths of 30bp are typically filtered, so the plot doesn't show these in many cases. A shifted distribution of read lengths towards smaller read lengths (e.g around 30-50bp) is also an indicator of ancient DNA.
                """,
                plot=self.lgdistplot(self.lgdist_rv_data, "Reverse"),
            )

    # Parse our nice little JSON file
    def parseJSON(self, f):
        """Parse the JSON output from DamageProfiler and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
        except Exception as e:
            print(e)
            log.warning(f"Could not parse DamageProfiler JSON: '{f['fn']}'")
            return None

        # Get sample name from JSON first
        s_name = self.clean_s_name(parsed_json["metadata"]["sample_name"], f)
        self.add_data_source(f, s_name)

        # Add version info
        version = parsed_json["metadata"]["version"]
        self.add_software_version(version, s_name)

        # Add 3' G to A data
        self.threepGtoAfreq_data[s_name] = parsed_json["dmg_3p"]

        # Add 5' C to T data
        self.fivepCtoTfreq_data[s_name] = parsed_json["dmg_5p"]

        # Add lendist forward
        self.lgdist_fw_data[s_name] = parsed_json["lendist_fw"]

        # Add lendist reverse
        self.lgdist_rv_data[s_name] = parsed_json["lendist_rv"]

        # Add summary metrics
        self.summary_metrics_data[s_name] = parsed_json["summary_stats"]

    #### Tables from here on

    def dmgprof_misinc_stats(self, dict_to_plot, readend, substitution):
        """Take the parsed stats from the DamageProfiler and add it to the
        basic stats table at the top of the report"""

        headers = {
            f"{readend}1": {
                "id": f"misinc-stats-1st-{readend}-{substitution}",
                "title": f"{readend} {substitution} 1st base",
                "description": f"{readend} 1st base substitution frequency for {substitution}",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGnBu",
                "modify": lambda x: x * 100.0,
            },
            f"{readend}2": {
                "id": f"misinc-stats-2nd-{readend}-{substitution}",
                "title": f"{readend} {substitution} 2nd base",
                "description": f"{readend} 2nd base substitution frequency for {substitution}",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "BuGn",
                "hidden": True,
                "modify": lambda x: x * 100.0,
            },
        }

        # Create new small subset dictionary for entries (we need just the first two data (k,v) pairs from each report)
        # Only the first two parts are informative from both 3' and 5' ends of reads (1st, 2nd base damage pattern)
        dict_to_add = dict()

        for key in dict_to_plot.keys():
            tmp = dict_to_plot[key]
            pos = [readend + "1", readend + "2"]
            strlist = tmp[:2]
            tuples = list(zip(pos, strlist))
            data = dict((x, y) for x, y in tuples)
            # Extract first two elements from list
            dict_to_add[key] = data

        self.general_stats_addcols(dict_to_add, headers)

    def add_summary_metrics(self, dict_to_plot):
        """Take the parsed stats from the DamageProfiler and add it to the
        basic stats table at the top of the report"""

        headers = {
            "std": {
                "title": "Read length std. dev.",
                "description": "Read length std. dev.",
                "suffix": "bp",
                "scale": "PuBu",
                "format": "{:,.2f}",
                "shared_key": "read_length",
                "hidden": True,
            },
            "median": {
                "title": "Median read length",
                "description": "Median read length",
                "suffix": "bp",
                "scale": "YlGnBu",
                "format": "{:,.2f}",
                "shared_key": "read_length",
            },
            "mean_readlength": {
                "title": "Mean read length",
                "description": "Mean read length",
                "suffix": "bp",
                "scale": "PuBuGn",
                "format": "{:,.2f}",
                "shared_key": "read_length",
                "hidden": True,
            },
        }

        self.general_stats_addcols(dict_to_plot, headers)

    #### Plotting from here on
    # Nice Linegraph plot for lgdist data

    def lgdistplot(self, dict_to_use, orientation):
        """Generate a read length distribution plot"""

        data = dict()
        for s_name in dict_to_use:
            try:
                data[s_name] = {int(d): int(dict_to_use[s_name][d]) for d in dict_to_use[s_name]}
            except KeyError:
                pass
        if len(data) == 0:
            log.debug("No valid data for forward read lgdist input!")
            return None

        config = {
            "id": f"length-distribution-{orientation}",
            "title": f"DamageProfiler: Read length distribution - {orientation} ",
            "ylab": "Number of reads",
            "xlab": "Readlength (bp)",
            "xDecimals": False,
            "tt_label": "{point.y} reads of length {point.x}",
            "ymin": 0,
            "xmin": 0,
        }
        return linegraph.plot(data, config)

    # Linegraph plot for 3pGtoA
    def threeprime_plot(self):
        """Generate a 3' G>A linegraph plot"""

        dict_to_add = dict()
        # Create tuples out of entries
        for key in self.threepGtoAfreq_data:
            pos = list(range(1, len(self.threepGtoAfreq_data.get(key))))
            # Multiply values by 100 to get %
            tmp = [i * 100.0 for i in self.threepGtoAfreq_data.get(key)]
            tuples = list(zip(pos, tmp))
            # Get a dictionary out of it
            data = dict((x, y) for x, y in tuples)
            dict_to_add[key] = data

        config = {
            "id": "threeprime_misinc_plot",
            "title": "DamageProfiler: 3' G>A misincorporation plot",
            "ylab": "% G to A substituted",
            "xlab": "Nucleotide position from 3'",
            "tt_label": "{point.y:.2f} % G>A misincorporations at nucleotide position {point.x}",
            "ymin": 0,
            "xmin": 1,
        }

        return linegraph.plot(dict_to_add, config)

    # Linegraph plot for 5pCtoT
    def fiveprime_plot(self):
        """Generate a 5' C>T linegraph plot"""

        dict_to_add = dict()
        # Create tuples out of entries
        for key in self.fivepCtoTfreq_data:
            pos = list(range(1, len(self.fivepCtoTfreq_data.get(key))))
            tmp = [i * 100.0 for i in self.fivepCtoTfreq_data.get(key)]
            tuples = list(zip(pos, tmp))
            # Get a dictionary out of it
            data = dict((x, y) for x, y in tuples)
            dict_to_add[key] = data

        config = {
            "id": "fiveprime_misinc_plot",
            "title": "DamageProfiler: 5' C>T misincorporation plot",
            "ylab": "% C to T substituted",
            "xlab": "Nucleotide position from 5'",
            "tt_label": "{point.y:.2f} % C>T misincorporations at nucleotide position {point.x}",
            "ymin": 0,
            "xmin": 1,
        }

        return linegraph.plot(dict_to_add, config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" __init__.py
~~~~~~~~~~~~~~~~~~~~
Initialises when multiqc module is loaded.

Makes the following available under the main multiqc namespace:
- run()
- config
- config.logger
- __version__
"""

import logging

from .multiqc import run
from .utils import config

config.logger = logging.getLogger(__name__)

__version__ = config.version

__all__ = ["run", "config", "__version__"]



""" MultiQC module to parse output from kraken """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, heatmap

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Kraken module"""

    def __init__(
        self,
        name="Kraken",
        anchor="kraken",
        href="https://ccb.jhu.edu/software/kraken/",
        info="is a taxonomic classification tool that uses exact k-mer matches to find the lowest common ancestor (LCA) of a given sequence.",
        doi="10.1186/gb-2014-15-3-r46",
        sp_key="kraken",
    ):
        super(MultiqcModule, self).__init__(
            name=name,
            anchor=anchor,
            href=href,
            info=info,
            doi=doi,
        )
        self.t_ranks = {
            "S": "Species",
            "G": "Genus",
            "F": "Family",
            "O": "Order",
            "C": "Class",
            "P": "Phylum",
            "K": "Kingdom",
            "D": "Domain",
            "R": "Root",
            "U": "Unclassified",
        }

        self.top_n = getattr(config, "kraken", {}).get("top_n", 5)

        # Find and load any kraken reports
        self.kraken_raw_data = dict()
        new_report_present = False
        for f in self.find_log_files(sp_key, filehandles=True):
            log_is_new = self.log_is_new(f)
            f["f"].seek(0)
            if not log_is_new:
                self.parse_logs(f)
            else:
                new_report_present = True
                self.parse_logs_minimizer(f)
            self.add_data_source(f)

        self.kraken_raw_data = self.ignore_samples(self.kraken_raw_data)

        if len(self.kraken_raw_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.kraken_raw_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.kraken_raw_data, f"multiqc_{self.anchor}")

        # Sum counts across all samples, so that we can pick top species
        self.kraken_total_pct = dict()
        self.kraken_total_counts = dict()
        self.kraken_sample_total_readcounts = dict()
        self.sample_total_readcounts()
        self.sum_sample_counts()

        self.general_stats_cols()
        self.top_taxa_barplot()
        if new_report_present:
            self.top_taxa_duplication_heatmap()

    @staticmethod
    def log_is_new(f):
        """Check which version of Kraken report file is used

        If 6 fields are used, it's the 'old' log (without distinct minimizer)
        if 8 fields, the new log experimental log (with distinct minimizer)
        """

        for line in f["f"]:
            if len(line.split()) > 6:
                return True
            else:
                return False
        return False

    def parse_logs(self, f):
        """
        Parses a kraken report output file

        1. Percentage of fragments covered by the clade rooted at this taxon
        2. Number of fragments covered by the clade rooted at this taxon
        3. Number of fragments assigned directly to this taxon
        4. A rank code, indicating:
            * (U)nclassified
            * (R)oot
            * (D)omain
            * (K)ingdom
            * (P)hylum
            * (C)lass
            * (O)rder
            * (F)amily
            * (G)enus
            * (S)pecies
           Taxa that are not at any of these 10 ranks have a rank code that is
           formed by using the rank code of the closest ancestor rank with
           a number indicating the distance from that rank.  E.g., "G2" is a
           rank code indicating a taxon is between genus and species and the
           grandparent taxon is at the genus rank.
        5. NCBI taxonomic ID number
        6. Indented scientific name
        """

        # Search regexes for stats
        k2_regex = re.compile(r"^\s{0,2}(\d{1,3}\.\d{1,2})\t(\d+)\t(\d+)\t([\dUDKRPCOFGS-]{1,3})\t(\d+)(\s+)(.+)")
        data = []
        for line in f["f"]:
            match = k2_regex.search(line)
            if match:
                row = {
                    "percent": float(match.group(1)),
                    "counts_rooted": int(match.group(2)),
                    "counts_direct": int(match.group(3)),
                    "rank_code": match.group(4),
                    "tax_id": int(match.group(5)),
                    "num_spaces": len(match.group(6)),
                    "classif": match.group(7),
                }
                data.append(row)

        self.kraken_raw_data[f["s_name"]] = data

    def parse_logs_minimizer(self, f):
        """
        Parses a kraken report output file

        1. Percentage of fragments covered by the clade rooted at this taxon
        2. Number of fragments covered by the clade rooted at this taxon
        3. Number of fragments assigned directly to this taxon
        4. Number of minimizers in read data associated with this taxon (new)
        5. An estimate of the number of distinct minimizers in read data
           associated with this taxon (new)
        6. A rank code, indicating:
            * (U)nclassified
            * (R)oot
            * (D)omain
            * (K)ingdom
            * (P)hylum
            * (C)lass
            * (O)rder
            * (F)amily
            * (G)enus
            * (S)pecies
           Taxa that are not at any of these 10 ranks have a rank code that is
           formed by using the rank code of the closest ancestor rank with
           a number indicating the distance from that rank.  E.g., "G2" is a
           rank code indicating a taxon is between genus and species and the
           grandparent taxon is at the genus rank.
        7. NCBI taxonomic ID number
        8. Indented scientific name
        """

        def duplication(total, distinct):
            """Get duplication factor

            Args:
                total (int): Total number of elements
                distinct (int): Number of distinct elements
            """
            try:
                res = float(total) / distinct
            except ZeroDivisionError:
                res = 0.0
            return res

        # Search regexes for stats
        k2_regex = re.compile(
            r"^\s{0,2}(\d{1,3}\.\d{1,2})\t(\d+)\t(\d+)\t(\d+)\t(\d+)\t([URDKPCOFGS-]\d{0,2})\t(\d+)(\s+)(.+)"
        )
        data = []
        for line in f["f"]:
            match = k2_regex.search(line)
            if match:
                row = {
                    "percent": float(match.group(1)),
                    "counts_rooted": int(match.group(2)),
                    "counts_direct": int(match.group(3)),
                    "minimizer": int(match.group(4)),
                    "minimizer_distinct": int(match.group(5)),
                    "minimizer_duplication": duplication(int(match.group(4)), int(match.group(5))),
                    "rank_code": match.group(6),
                    "tax_id": int(match.group(7)),
                    "num_spaces": len(match.group(8)),
                    "classif": match.group(9),
                }
                data.append(row)
            else:
                log.debug(f"{f['s_name']}: Could not parse line: {line}")

        self.kraken_raw_data[f["s_name"]] = data

    def sample_total_readcounts(self):
        """Compute the total read counts for each sample"""

        total_all_samples = 0
        for s_name, data in self.kraken_raw_data.items():
            self.kraken_sample_total_readcounts[s_name] = 0
            for row in data:
                self.kraken_sample_total_readcounts[s_name] += row["counts_direct"]
            total_all_samples += self.kraken_sample_total_readcounts[s_name]

        # Check that we had some counts for some samples, exit if not
        if total_all_samples == 0:
            log.warning("No samples had any reads")
            raise ModuleNoSamplesFound

    def sum_sample_counts(self):
        """Sum counts across all samples for kraken data"""

        # Sum the percentages for each taxa across all samples
        # Allows us to pick the top taxa for each rank
        # Use percentages instead of counts so that deeply-sequences samples
        # are not unfairly over-represented
        for s_name, data in self.kraken_raw_data.items():
            for row in data:
                # Convenience vars that are easier to read
                rank_code = row["rank_code"]
                classif = row["classif"]

                # Skip anything that doesn't exactly fit a tax rank level
                if row["rank_code"] == "-" or any(c.isdigit() for c in row["rank_code"]):
                    continue

                if rank_code not in self.kraken_total_pct:
                    self.kraken_total_pct[rank_code] = dict()
                    self.kraken_total_counts[rank_code] = dict()

                if classif not in self.kraken_total_pct[rank_code]:
                    self.kraken_total_pct[rank_code][classif] = 0
                    self.kraken_total_counts[rank_code][classif] = 0
                try:
                    self.kraken_total_pct[rank_code][classif] += (
                        row["counts_rooted"] / self.kraken_sample_total_readcounts[s_name]
                    )
                except ZeroDivisionError:
                    pass
                self.kraken_total_counts[rank_code][classif] += row["counts_rooted"]

    def general_stats_cols(self):
        """Add a couple of columns to the General Statistics table"""

        # Get top taxa in most specific taxa rank that we have
        top_taxa = []
        top_rank_code = None
        top_rank_name = None
        for rank_code, rank_name in self.t_ranks.items():
            try:
                sorted_pct = sorted(self.kraken_total_pct[rank_code].items(), key=lambda x: x[1], reverse=True)
                for classif, pct_sum in sorted_pct[: self.top_n]:
                    top_taxa.append(classif)
                top_rank_code = rank_code
                top_rank_name = rank_name
                break
            except KeyError:
                # No species-level data found etc
                pass

        # Column headers
        headers = dict()

        top_one = None

        # don't include top-N % in general stats if all is unclassified.
        # unclassified is included separately, so also don't include twice
        if top_rank_code != "U":
            top_one = f"% {top_taxa[0]}"
            headers["pct_top_one"] = {
                "title": top_one,
                "description": "Percentage of reads that were the top {} over all samples - {}".format(
                    top_rank_name, top_taxa[0]
                ),
                "suffix": "%",
                "max": 100,
                "scale": "PuBuGn",
            }
            headers["pct_top_n"] = {
                "title": f"% Top {self.top_n} {top_rank_name}",
                "description": f"Percentage of reads that were classified by one of the top-{self.top_n} {top_rank_name} ({', '.join(top_taxa)})",
                "suffix": "%",
                "max": 100,
                "scale": "PuBu",
            }

        headers["pct_unclassified"] = {
            "title": "% Unclassified",
            "description": "Percentage of reads that were unclassified",
            "suffix": "%",
            "max": 100,
            "scale": "OrRd",
        }

        # Get table data
        tdata = {}
        for s_name, d in self.kraken_raw_data.items():
            tdata[s_name] = {}
            for row in d:
                try:
                    percent = (row["counts_rooted"] / self.kraken_sample_total_readcounts[s_name]) * 100.0
                except ZeroDivisionError:
                    percent = 0
                if row["rank_code"] == "U":
                    tdata[s_name]["pct_unclassified"] = percent
                if row["rank_code"] == top_rank_code and row["classif"] in top_taxa:
                    tdata[s_name]["pct_top_n"] = percent + tdata[s_name].get("pct_top_n", 0)
                if row["rank_code"] == top_rank_code and row["classif"] == top_taxa[0]:
                    tdata[s_name]["pct_top_one"] = percent

            if top_one is not None and "pct_top_one" not in tdata[s_name]:
                tdata[s_name]["pct_top_one"] = 0

        self.general_stats_addcols(tdata, headers)

    def top_taxa_barplot(self):
        """Add a bar plot showing the top-N from each taxa rank"""

        pd = []
        cats = list()
        # Keeping track of encountered codes to display only tabs with available data
        found_rank_codes = set()

        for rank_code in self.t_ranks:
            rank_cats = dict()
            rank_data = dict()

            # Loop through the summed tax percentages to get the top-N across all samples
            try:
                sorted_pct = sorted(self.kraken_total_pct[rank_code].items(), key=lambda x: x[1], reverse=True)
            except KeyError:
                # Taxa rank not found in this sample
                continue
            i = 0
            counts_shown = {}
            for classif, pct_sum in sorted_pct:
                i += 1
                if i > self.top_n:
                    break
                rank_cats[classif] = {"name": classif}
                # Pull out counts for this rank + classif from each sample
                for s_name, d in self.kraken_raw_data.items():
                    if s_name not in rank_data:
                        rank_data[s_name] = dict()
                    if s_name not in counts_shown:
                        counts_shown[s_name] = 0

                    for row in d:
                        if row["rank_code"] == rank_code:
                            found_rank_codes.add(rank_code)
                            # unclassified are handled separately
                            if row["rank_code"] != "U":
                                if row["classif"] == classif:
                                    if classif not in rank_data[s_name]:
                                        rank_data[s_name][classif] = 0
                                    rank_data[s_name][classif] += row["counts_rooted"]
                                    counts_shown[s_name] += row["counts_rooted"]

            # Add in unclassified reads and "other" - we presume from other species etc.
            for s_name, d in self.kraken_raw_data.items():
                for row in d:
                    if row["rank_code"] == "U":
                        rank_data[s_name]["U"] = row["counts_rooted"]
                        counts_shown[s_name] += row["counts_rooted"]
                rank_data[s_name]["other"] = self.kraken_sample_total_readcounts[s_name] - counts_shown[s_name]

                # This should never happen... But it does sometimes if the total read count is a bit off
                if rank_data[s_name]["other"] < 0:
                    log.debug(
                        "Found negative 'other' count for {} ({}): {}".format(
                            s_name, self.t_ranks[rank_code], rank_data[s_name]["other"]
                        )
                    )
                    rank_data[s_name]["other"] = 0

            rank_cats["other"] = {"name": "Other", "color": "#cccccc"}
            rank_cats["U"] = {"name": "Unclassified", "color": "#d4949c"}

            cats.append(rank_cats)
            pd.append(rank_data)

        pconfig = {
            "id": f"{self.anchor}-top-n-plot",
            "title": f"{self.name}: Top taxa",
            "ylab": "Number of fragments",
            "data_labels": [v for k, v in self.t_ranks.items() if k in found_rank_codes],
        }

        self.add_section(
            name="Top taxa",
            anchor=f"{self.anchor}-top-n",
            description=f"The number of reads falling into the top {self.top_n} taxa across different ranks.",
            helptext=f"""
                To make this plot, the percentage of each sample assigned to a given taxa is summed across all samples.
                The counts for these top {self.top_n} taxa are then plotted for each of the 9 different taxa ranks.
                The unclassified count is always shown across all taxa ranks.

                The total number of reads is approximated by dividing the number of `unclassified` reads by the percentage of
                the library that they account for.
                Note that this is only an approximation, and that kraken percentages don't always add to exactly 100%.

                The category _"Other"_ shows the difference between the above total read count and the sum of the read counts
                in the top {self.top_n} taxa shown + unclassified. This should cover all taxa _not_ in the top {self.top_n}, +/- any rounding errors.

                Note that any taxon that does not exactly fit a taxon rank (eg. `-` or `G2`) is ignored.
            """,
            plot=bargraph.plot(pd, cats, pconfig),
        )

    def top_taxa_duplication_heatmap(self):
        """Add a heatmap showing the minimizer duplication of the top taxa"""

        duplication = list()
        pconfig = {
            "id": f"{self.anchor}-top-duplication_plot",
            "title": f"{self.name}: Top {self.top_n} species duplication",
            "square": False,
            "xcats_samples": False,
            "angled_xticks": False,
        }

        rank_code = "S"
        rank_data = dict()
        # Loop through the summed tax percentages to get the top taxa across all samples
        try:
            sorted_pct = sorted(self.kraken_total_pct[rank_code].items(), key=lambda x: x[1], reverse=True)
        except KeyError:
            log.debug("Taxa rank not found, skipping Taxa duplication heatmap")
            return

        i = 0
        counts_shown = {}
        showed_warning = False
        for classif, pct_sum in sorted_pct:
            i += 1
            if i > self.top_n:
                break
            # Pull out counts for this rank + classif from each sample
            for s_name, d in self.kraken_raw_data.items():
                if s_name not in rank_data:
                    rank_data[s_name] = dict()
                if s_name not in counts_shown:
                    counts_shown[s_name] = 0

                if classif not in rank_data[s_name]:
                    rank_data[s_name][classif] = None

                try:
                    row = next(row for row in d if row["rank_code"] == rank_code and row["classif"] == classif)
                except StopIteration:
                    # if nothing is found at the rank + classification, leave as 0
                    continue

                try:
                    rank_data[s_name][classif] = row["minimizer_duplication"]
                except KeyError:
                    del rank_data[s_name]
                    if not showed_warning:
                        log.warning("Kraken2 reports of different versions were found")
                        showed_warning = True

        # Strip empty samples
        for sample, vals in dict(rank_data).items():
            if len(vals) == 0:
                del rank_data[sample]

        # Build data structures for heatmap
        ylabels = list(rank_data.keys())
        xlabels = list(rank_data[ylabels[0]].keys())
        for sample in rank_data:
            duplication.append(list(rank_data[sample].values()))

        self.add_section(
            name="Duplication rate of top species",
            anchor=f"{self.anchor}-duplication-topfive",
            description=f"The duplication rate of minimizer falling into the top {self.top_n} species",
            helptext=f"""
                To make this plot, the minimizer duplication rate is computed for the top {self.top_n} most abundant species in all samples.

                The minimizer duplication rate is defined as: `duplication rate = (total number of minimizers / number of distinct minimizers)`

                A low coverage and high duplication rate (`>> 1`) is often sign of read stacking, which probably indicates of false positive hit.
            """,
            plot=heatmap.plot(duplication, xlabels, ylabels, pconfig),
        )



#!/usr/bin/env python3

import logging
import re
import traceback

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph

# Initialize log
log = logging.getLogger(__name__)

VERSION_REGEX = r"\[FLASH\] FLASH v([\d\.]+) complete!"


class MultiqcModule(BaseMultiqcModule):
    """FLASh MultiQC module

    Options:
      use_output_name: true - use first output filename as sample name
        default uses first input filename in log
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="FLASh",
            anchor="flash",
            href="https://ccb.jhu.edu/software/FLASH/",
            info="is a very fast and accurate software tool to merge paired-end reads from next-generation sequencing experiments.",
            doi="10.1093/bioinformatics/btr507",
        )

        flash_results = self.parse_flash()

        hist_results = self.hist_results()

        if not flash_results and not hist_results:
            raise ModuleNoSamplesFound

    def parse_flash(self):
        # Find all log files with flash msgs
        flash_data = {}
        for logfile in self.find_log_files("flash/log"):
            flash_data.update(self.parse_flash_log(logfile))
            self.add_data_source(logfile)

        # Ignore sample names
        flash_data = self.ignore_samples(flash_data)
        if not flash_data:
            return 0

        log.info(f"Found {len(flash_data)} log reports")

        self.general_stats_table(flash_data)
        self.add_section(
            name="Read combination statistics", anchor="flash-bargraph", plot=self.summary_plot(flash_data)
        )

        self.write_data_file(flash_data, "multiqc_flash_combo_stats")
        return len(flash_data)

    @staticmethod
    def split_log(logf):
        """split concat log into individual samples"""
        flashpatt = re.compile(
            r"\[FLASH\] Fast Length Adjustment of SHort reads\n(.+?)\[FLASH\] FLASH", flags=re.DOTALL
        )
        return flashpatt.findall(logf)

    @staticmethod
    def get_field(field, slog, fl=False):
        """parse sample log for field
        set fl=True to return a float
        otherwise, returns int
        """
        field += r"\:\s+([\d\.]+)"
        match = re.search(field, slog)
        if match:
            if fl:
                return float(match.group(1))
            return int(match.group(1))
        return 0

    def clean_pe_name(self, nlog, logf):
        """additional name cleaning for paired end data"""
        use_output_name = getattr(config, "flash", {}).get("use_output_name", False)
        if use_output_name:
            name = re.search(r"Output files\:\n\[FLASH\]\s+(.+?)\n", nlog)
        else:
            name = re.search(r"Input files\:\n\[FLASH\]\s+(.+?)\n", nlog)
        if not name:
            return None
        name = name.group(1)
        name = self.clean_s_name(name, logf)
        return name

    def parse_flash_log(self, logf):
        """parse flash logs"""
        data = {}
        samplelogs = self.split_log(logf["f"])
        version_match = re.search(VERSION_REGEX, logf["f"])

        for slog in samplelogs:
            try:
                sample = dict()
                ## Sample name ##
                s_name = self.clean_pe_name(slog, logf)
                if s_name is None:
                    continue
                sample["s_name"] = s_name

                if version_match:
                    self.add_software_version(version_match.group(1), s_name)

                ## Log attributes ##
                sample["totalpairs"] = self.get_field("Total pairs", slog)
                sample["discardpairs"] = self.get_field("Discarded pairs", slog)
                sample["percdiscard"] = self.get_field("Percent Discarded", slog, fl=True)
                sample["combopairs"] = self.get_field("Combined pairs", slog)
                sample["inniepairs"] = self.get_field("Innie pairs", slog)
                sample["outiepairs"] = self.get_field("Outie pairs", slog)
                sample["uncombopairs"] = self.get_field("Uncombined pairs", slog)
                sample["perccombo"] = self.get_field("Percent combined", slog, fl=True)

                data[s_name] = sample
            except Exception as err:
                log.warning(f"Error parsing record in {logf['fn']}. {err}")
                log.debug(traceback.format_exc())
                continue
        return data

    def general_stats_table(self, data):
        """Add percent combined to general stats table"""
        headers = {
            "combopairs": {
                "title": "Combined pairs",
                "description": "Num read pairs combined",
                "shared_key": "read_count",
                "hidden": True,
                "scale": False,
            },
            "perccombo": {
                "title": "% Combined",
                "description": "% read pairs combined",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "PiYG",
            },
        }
        self.general_stats_addcols(data, headers)

    @staticmethod
    def summary_plot(data):
        """Barplot of combined pairs"""
        cats = {
            "combopairs": {"name": "Combined pairs", "color": "#191970"},
            "inniepairs": {"name": "Combined innie pairs", "color": "#191970"},
            "outiepairs": {"name": "Combined outie pairs", "color": "#00A08A"},
            "uncombopairs": {"name": "Uncombined pairs", "color": "#cd1076"},
            "discardpairs": {"name": "Discarded pairs", "color": "#ffd700"},
        }
        splotconfig = {
            "id": "flash_combo_stats_plot",
            "title": "FLASh: Read combination statistics",
            "ylab": "Number of read pairs",
        }
        # Only plot the combopairs category if we don't have inniepairs and outiepairs
        for s_name, d in data.items():
            if (d["inniepairs"] > 0 or d["outiepairs"] > 0) and d["combopairs"] > 0:
                del data[s_name]["combopairs"]

        return bargraph.plot(data, cats, splotconfig)

    @staticmethod
    def parse_hist_files(histf):
        """parse histogram files"""
        nameddata = dict()
        data = dict()
        try:
            for line in histf["f"].splitlines():
                s = line.split()
                if s:
                    if len(s) != 2:
                        raise RuntimeError(
                            "invalid format: " + str(len(s)) + " column(s) found in row. must be exactly 2."
                        )
                    data[int(s[0])] = int(s[1])
        except Exception as err:
            log.warning("Error parsing %s. %s", histf["fn"], err)
            log.debug(traceback.format_exc())
        else:
            if data:
                nameddata[histf["s_name"]] = data
            else:
                log.debug("%s is empty.", histf["fn"])
        finally:
            return nameddata

    @staticmethod
    def get_colors(n):
        """get colors for freqpoly graph"""
        cb_palette = [
            "#E69F00",
            "#56B4E9",
            "#009E73",
            "#F0E442",
            "#0072B2",
            "#D55E00",
            "#CC79A7",
            "#001F3F",
            "#0074D9",
            "#7FDBFF",
            "#39CCCC",
            "#3D9970",
            "#2ECC40",
            "#01FF70",
            "#FFDC00",
            "#FF851B",
            "#FF4136",
            "#F012BE",
            "#B10DC9",
            "#85144B",
            "#AAAAAA",
            "#000000",
        ]

        whole = int(n / 22)
        extra = n % 22
        cols = cb_palette * whole
        if extra >= 0:
            cols.extend(cb_palette[0:extra])
        return cols

    @staticmethod
    def freqpoly_plot(data):
        """make freqpoly plot of merged read lengths"""
        rel_data = {}
        for key, val in data.items():
            tot = sum(val.values(), 0)
            rel_data[key] = {k: v / tot for k, v in val.items()}
        fplotconfig = {
            "id": "flash_freqpoly_plot",
            "title": "FLASh: Frequency of merged read lengths",
            "xlab": "Merged Read Length",
            "ylab": "Frequency",
            "data_labels": [
                {"name": "Absolute", "ylab": "Frequency", "xlab": "Merged Read Length"},
                {"name": "Relative", "ylab": "Relative Frequency", "xlab": "Merged Read Length"},
            ],
            "colors": dict(zip(data.keys(), MultiqcModule.get_colors(len(data)))),
        }
        return linegraph.plot([data, rel_data], fplotconfig)

    def hist_results(self):
        """process flash numeric histograms"""
        hist_data = {}
        for f in self.find_log_files("flash/hist"):
            hist_data.update(self.parse_hist_files(f))

        # ignore sample names
        hist_data = self.ignore_samples(hist_data)
        if not hist_data:
            return 0

        log.info("Found %d histogram reports", len(hist_data))

        self.add_section(
            name="Frequency polygons of merged read lengths",
            anchor="flash-histogram",
            description="This plot is made from the numerical histograms output by FLASh.",
            plot=self.freqpoly_plot(hist_data),
        )
        return len(hist_data)



""" MultiQC submodule to parse output from RSeQC read_GC.py
http://rseqc.sourceforge.net/#read-gc-py """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC read_GC reports and parse their data"""

    # Set up vars
    self.read_gc = dict()
    self.read_gc_pct = dict()

    # Go through files and parse data
    for f in self.find_log_files("rseqc/read_gc"):
        if f["f"].startswith("GC%	read_count"):
            gc = list()
            counts = list()
            for line in f["f"].splitlines():
                s = line.split()
                try:
                    gc.append(float(s[0]))
                    counts.append(float(s[1]))
                except Exception:
                    pass
            if len(gc) > 0:
                sorted_gc_keys = sorted(range(len(gc)), key=lambda k: gc[k])
                total = sum(counts)
                if f["s_name"] in self.read_gc:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="read_GC")
                self.read_gc[f["s_name"]] = dict()
                self.read_gc_pct[f["s_name"]] = dict()
                for i in sorted_gc_keys:
                    self.read_gc[f["s_name"]][gc[i]] = counts[i]
                    self.read_gc_pct[f["s_name"]][gc[i]] = (counts[i] / total) * 100

    # Filter to strip out ignored sample names
    self.read_gc = self.ignore_samples(self.read_gc)

    if len(self.read_gc) == 0:
        return 0

    # Write data to file
    self.write_data_file(self.read_gc, "rseqc_read_gc")

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Add line graph to section
    pconfig = {
        "id": "rseqc_read_gc_plot",
        "title": "RSeQC: Read GC Content",
        "ylab": "Number of Reads",
        "xlab": "GC content (%)",
        "xmin": 0,
        "xmax": 100,
        "tt_label": "<strong>{point.x}% GC</strong>: {point.y:.2f}",
        "data_labels": [
            {"name": "Counts", "ylab": "Number of Reads"},
            {"name": "Percentages", "ylab": "Percentage of Reads"},
        ],
    }
    self.add_section(
        name="Read GC Content",
        anchor="rseqc-read_gc",
        description='<a href="http://rseqc.sourceforge.net/#read-gc-py" target="_blank">read_GC</a>'
        " calculates a histogram of read GC content.</p>",
        plot=linegraph.plot([self.read_gc, self.read_gc_pct], pconfig),
    )

    # Return number of samples found
    return len(self.read_gc)



""" MultiQC module to parse output from Bakta """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bakta",
            anchor="bakta",
            href="https://github.com/oschwengers/bakta",
            info="is a tool for the rapid & standardized annotation of bacterial genomes, MAGs & plasmids",
            doi="10.1099/mgen.0.000685",
        )

        # Parse logs (txt files)
        self.bakta = dict()
        for f in self.find_log_files("bakta"):
            self.bakta[f["s_name"]] = self.parse_bakta(f)
            self.add_data_source(f)
        # Filter out ignored samples (given with --ignore-samples option)
        self.bakta = self.ignore_samples(self.bakta)

        # Raise error if dict is empty
        if len(self.bakta) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.bakta)} logs")

        # Write parsed report data to a file
        self.write_data_file(self.bakta, "multiqc_bakta")

        # Add most important Bakta annotation stats to general table
        headers = {
            "Count": {
                "title": "# contigs",
                "description": "Number of contigs",
                "min": 0,
                "scale": "Blues",
                "format": "{:,d}",
            },
            "Length": {
                "title": "# bases",
                "description": "Total number of bases in the contigs",
                "min": 0,
                "scale": "YlGn",
                "format": "{:,d}",
            },
            "CDSs": {
                "title": "# CDS",
                "description": "Number of found CDS",
                "min": 0,
                "scale": "YlGnBu",
                "format": "{:,d}",
                "shared_key": "gene_count",
            },
        }
        self.general_stats_addcols(self.bakta, headers)

        # Helptext for Bakta barplot:
        descr_plot = "This barplot shows the distribution of different types of features found in each sample."

        helptext = """
           `Bakta` can detect different features:

            - tRNAs
            - tmRNAs
            - rRNAs
            - ncRNAs
            - ncRNA regions
            - CRISPR arrays
            - CDSs
            - pseudogenes
            - hypotheticals
            - signal peptides
            - sORFs
            - gaps
            - oriCs
            - oriVs
            - oriTs

            This barplot shows you the distribution of these different types of features found in each sample.
            """
        self.add_section(plot=self.bakta_barplot(), helptext=helptext, description=descr_plot)

    def parse_bakta(self, f):
        """Parse bakta txt summary files for annotation parameters."""

        metrics = (
            "Length",
            "Count",
            "N50",
            "tRNAs",
            "tmRNAs",
            "rRNAs",
            "ncRNAs",
            "ncRNA regions",
            "CRISPR arrays",
            "CDSs",
            "pseudogenes",
            "hypotheticals",
            "signal peptides",
            "sORFs",
            "gaps",
            "oriCs",
            "oriVs",
            "oriTs",
        )

        data = {}
        for line in f["contents_lines"]:
            s = line.strip().split(": ")
            if s[0] in metrics:
                data[s[0].replace(" ", "_")] = int(s[1])

            if s[0] == "Software":
                self.add_software_version(s[1].strip("v"), f["s_name"])
        return data

    def bakta_barplot(self):
        """Make a basic plot of the annotation stats"""

        # Specify the order of the different categories
        categories = {
            metric: {"name": cat}
            for metric, cat in {
                "CDSs": "CDS",
                "rRNAs": "rRNA",
                "tRNAs": "tRNA",
                "tmRNAs": "tmRNA",
                "ncRNAs": "ncRNA",
                "ncRNA_regions": "ncRNA region",
                "hypotheticals": "hypothetical",
                "CRISPR_arrays": "CRISPR array",
                "pseudogenes": "pseudogene",
                "signal_peptides": "signal peptide",
                "sORFs": "sORF",
                "gaps": "gap",
                "oriCs": "oriC",
                "oriVs": "oriV",
                "oriTs": "oriT",
            }.items()
        }

        plot_config = {
            "id": "bakta_feature_types",
            "title": "Bakta: Feature Types",
            "ylab": "# Contigs",
            "cpswitch_counts_label": "Features",
        }
        return bargraph.plot(self.bakta, categories, plot_config)



""" MultiQC functions to plot a box plot """

from typing import List, Dict, Union, OrderedDict

import inspect
import logging
import re

from multiqc.plots.plotly.box import BoxT
from multiqc.utils import config, report
from multiqc.plots.plotly import box

logger = logging.getLogger(__name__)

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(list_of_data_by_sample: Union[Dict[str, BoxT], List[Dict[str, BoxT]]], pconfig=None):
    """
    Plot a box plot. Expects either:
    - a dict mapping sample names to data point lists or dicts,
    - a dict mapping sample names to a dict of statistics (e.g. {min, max, median, mean, std, q1, q3 etc.})
    """
    if pconfig is None:
        pconfig = {}

    # Given one dataset - turn it into a list
    if not isinstance(list_of_data_by_sample, list):
        list_of_data_by_sample = [list_of_data_by_sample]

    for i in range(len(list_of_data_by_sample)):
        if isinstance(list_of_data_by_sample[0], OrderedDict):
            # Legacy: users assumed that passing an OrderedDict indicates that we
            # want to keep the sample order https://github.com/MultiQC/MultiQC/issues/2204
            pass
        elif pconfig.get("sort_samples", True):
            samples = sorted(list(list_of_data_by_sample[0].keys()))
            list_of_data_by_sample[i] = {s: list_of_data_by_sample[i][s] for s in samples}

    # Allow user to overwrite any given config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for k, v in config.custom_plot_config[pconfig["id"]].items():
            pconfig[k] = v

    # Validate config if linting
    if config.strict:
        # Get module name
        modname = ""
        callstack = inspect.stack()
        for n in callstack:
            if "multiqc/modules/" in n[1] and "base_module.py" not in n[1]:
                callpath = n[1].split("multiqc/modules/", 1)[-1]
                modname = f">{callpath}< "
                break
        # Look for essential missing pconfig keys
        for k in ["id", "title"]:
            if k not in pconfig:
                errmsg = f"LINT: {modname}Box plot pconfig was missing key '{k}'"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)
        # Check plot title format
        if not re.match(r"^[^:]*\S: \S[^:]*$", pconfig.get("title", "")):
            errmsg = "LINT: {} Box plot title did not match format 'Module: Plot Name' (found '{}')".format(
                modname, pconfig.get("title", "")
            )
            logger.error(errmsg)
            report.lint_errors.append(errmsg)

    # Make a plot - custom, interactive or flat
    mod = get_template_mod()
    if "box" in mod.__dict__ and callable(mod.bargraph):
        try:
            return mod.box(list_of_data_by_sample, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return box.plot(list_of_data_by_sample, pconfig)



""" MultiQC module to parse output from Lima """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import HUMID submodules
from . import stats, neighbours, counts, clusters

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialse the parent object
        super(MultiqcModule, self).__init__(
            name="HUMID",
            anchor="humid",
            href="https://github.com/jfjlaros/HUMID",
            info=" is a fast, reference free tool to remove (UMI) duplicates from sequencing data",
            # No publication / DOI // doi=
        )
        self.stats = None
        self.neighbours = None
        self.counts = None
        self.clusters = None

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Look for stats files
        stats.parse_reports(self)

        # Look for neighbour files
        neighbours.parse_reports(self)

        # Look for count files
        counts.parse_reports(self)

        # Look for cluster files
        clusters.parse_reports(self)

        if all(not x for x in [self.stats, self.neighbours, self.counts, self.clusters]):
            raise ModuleNoSamplesFound

        num_samples = max([len(x) for x in [self.stats, self.neighbours, self.counts, self.clusters]])
        log.info(f"Found {num_samples} reports")



""" MultiQC module to parse output from Cell Ranger count """

import json
import logging
import re

from multiqc import config
from multiqc.modules.cellranger.utils import set_hidden_cols, update_dict, parse_bcknee_data, clean_title_case
from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class CellRangerVdjMixin:
    """Cell Ranger count report parser"""

    def parse_vdj_html(self):
        self.cellrangervdj_mapping = dict()
        self.cellrangervdj_annotations = dict()
        self.cellrangervdj_general_data = dict()
        self.cellrangervdj_warnings = dict()
        self.cellrangervdj_plots_conf = {"bc": dict(), "genes": dict()}
        self.cellrangervdj_plots_data = {"bc": dict(), "genes": dict()}
        self.vdj_general_data_headers = dict()
        self.vdj_mapping_headers = dict()
        self.vdj_annotations_headers = dict()
        self.vdj_warnings_headers = dict()

        for f in self.find_log_files("cellranger/vdj_html", filehandles=True):
            self.parse_vdj_report(f)

        self.cellrangervdj_mapping = self.ignore_samples(self.cellrangervdj_mapping)
        self.cellrangervdj_annotations = self.ignore_samples(self.cellrangervdj_annotations)
        self.cellrangervdj_general_data = self.ignore_samples(self.cellrangervdj_general_data)
        self.cellrangervdj_warnings = self.ignore_samples(self.cellrangervdj_warnings)
        for k in self.cellrangervdj_plots_data.keys():
            self.cellrangervdj_plots_data[k] = self.ignore_samples(self.cellrangervdj_plots_data[k])

        self.vdj_general_data_headers["reads"] = {
            "title": f"{config.read_count_prefix} Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "namespace": "VDJ",
        }
        self.vdj_general_data_headers = set_hidden_cols(
            self.vdj_general_data_headers,
            [
                "Q30 bc",
                "Q30 UMI",
                "Q30 read",
                "Q30 read1",
                "Q30 read2",
            ],
        )

        self.vdj_mapping_headers["reads"] = {
            "title": f"{config.read_count_prefix} Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "modify": lambda x: x * config.read_count_multiplier,
        }
        self.vdj_mapping_headers = set_hidden_cols(
            self.vdj_mapping_headers,
            [
                "Q30 bc",
                "Q30 UMI",
                "Q30 read1",
                "Q30 read2",
                "IGH reads",
                "IGK reads",
                "IGL reads",
            ],
        )

        self.vdj_annotations_headers = set_hidden_cols(
            self.vdj_annotations_headers,
            [
                "cells IGH contig",
                "cells IGK contig",
                "cells IGL contig",
                "cells IGH CDR3",
                "cells IGK CDR3",
                "cells IGL CDR3",
                "cells IGH VJ span",
                "cells IGK VJ span",
                "cells IGL VJ span",
            ],
        )

        if len(self.cellrangervdj_general_data) == 0:
            return 0

        else:
            for k in self.vdj_general_data_headers.keys():
                self.vdj_general_data_headers[k]["title"] = f"{self.vdj_general_data_headers[k]['title']} (VDJ)"
                self.vdj_general_data_headers[k][
                    "description"
                ] = f"{self.vdj_general_data_headers[k]['description']} (VDJ)"

            self.general_stats_addcols(self.cellrangervdj_general_data, self.vdj_general_data_headers)

            # Write parsed report data to a file
            self.write_data_file(self.cellrangervdj_mapping, "multiqc_cellranger_vdj_mapping")
            self.write_data_file(self.cellrangervdj_annotations, "multiqc_cellranger_vdj_annotations")

            # Add sections to the report
            if len(self.cellrangervdj_warnings) > 0:
                self.add_section(
                    name="VDJ - Warnings",
                    anchor="cellranger-vdj-warnings",
                    description="Warnings encountered during the analysis",
                    plot=table.plot(
                        self.cellrangervdj_warnings,
                        self.vdj_warnings_headers,
                        {
                            "namespace": "VDJ",
                            "id": "cellranger-vdj-warnings-table",
                            "title": "Cellranger VDJ: Warnings",
                        },
                    ),
                )

            self.add_section(
                name="VDJ - Summary stats",
                anchor="cellranger-vdj-stats",
                description="Summary QC metrics from Cell Ranger count",
                plot=table.plot(
                    self.cellrangervdj_mapping,
                    self.vdj_mapping_headers,
                    {
                        "namespace": "VDJ",
                        "id": "cellranger-vdj-stats-table",
                        "title": "Cellranger VDJ: Summary stats",
                    },
                ),
            )

            self.add_section(
                name="VDJ - Annotations",
                anchor="cellranger-vdj-annot",
                description="V(D)J annotations from Cell Ranger VDJ analysis",
                plot=table.plot(
                    self.cellrangervdj_annotations,
                    self.vdj_annotations_headers,
                    {
                        "namespace": "VDJ",
                        "id": "cellranger-vdj-annot-table",
                        "title": "Cellranger VDJ: Annotations",
                    },
                ),
            )

            self.add_section(
                name="VDJ - BC rank plot",
                anchor="cellranger-vdj-bcrank-plot",
                description=self.cellrangervdj_plots_conf["bc"]["description"],
                helptext=self.cellrangervdj_plots_conf["bc"]["helptext"],
                plot=linegraph.plot(self.cellrangervdj_plots_data["bc"], self.cellrangervdj_plots_conf["bc"]["config"]),
            )

            return len(self.cellrangervdj_general_data)

    def parse_vdj_report(self, f):
        """Go through the html report of cell ranger and extract the data in a dicts"""

        for line in f["f"]:
            line = line.strip()
            if line.startswith("const data"):
                line = line.replace("const data = ", "")
                mydict = json.loads(line)
                mydict = mydict["summary"]
                break

        s_name = self.clean_s_name(mydict["sample"]["id"], f)

        # Extract software version
        try:
            version_pair = mydict["summary_tab"]["pipeline_info_table"]["rows"][-1]
            assert version_pair[0] == "Pipeline Version"
            version_match = re.search(r"cellranger-([\d\.]+)", version_pair[1])
            if version_match:
                self.add_software_version(version_match.group(1), s_name)
        except (KeyError, AssertionError):
            log.debug(f"Unable to parse version for sample {s_name}")

        data_general_stats = dict()

        # Store general stats from cells and sequencing tables
        col_dict = {
            "Estimated Number of Cells": "estimated cells",
            "Mean Reads per Cell": "avg reads/cell",
            "Fraction Reads in Cells": "reads in cells",
        }
        colours = {
            "estimated cells": "RdPu",
            "avg reads/cell": "Blues",
            "reads in cells": "PuBuGn",
        }
        data_general_stats, self.vdj_general_data_headers = update_dict(
            data_general_stats,
            self.vdj_general_data_headers,
            mydict["summary_tab"]["cells"]["table"]["rows"],
            col_dict,
            colours,
            "VDJ",
        )

        col_dict = {
            "Number of Reads": "reads",
            "Valid Barcodes": "valid bc",
            "Q30 Bases in Barcode": "Q30 bc",
            "Q30 Bases in UMI": "Q30 UMI",
            "Q30 Bases in RNA Read 1": "Q30 read1",
            "Q30 Bases in RNA Read 2": "Q30 read2",
        }
        data_general_stats, self.vdj_general_data_headers = update_dict(
            data_general_stats,
            self.vdj_general_data_headers,
            mydict["summary_tab"]["vdj_sequencing"]["table"]["rows"],
            col_dict,
            {},
            "VDJ",
        )

        # Store sequencing results data from vdj report
        data_rows = (
            mydict["summary_tab"]["vdj_sequencing"]["table"]["rows"]
            + mydict["summary_tab"]["cells"]["table"]["rows"]
            + mydict["summary_tab"]["vdj_enrichment"]["table"]["rows"]
        )
        col_dict = {
            "Number of Reads": "reads",
            "Valid Barcodes": "valid bc",
            "Estimated Number of Cells": "estimated cells",
            "Mean Reads per Cell": "avg reads/cell",
            "Mean Used Reads per Cell": "used reads/cell",
            "Fraction Reads in Cells": "reads in cells",
            "Reads Mapped to Any V(D)J Gene": "VDJ reads",
            "Reads Mapped to IGH": "IGH reads",
            "Reads Mapped to IGK": "IGK reads",
            "Reads Mapped to IGL": "IGL reads",
            "Q30 Bases in Barcode": "Q30 bc",
            "Q30 Bases in UMI": "Q30 UMI",
            "Q30 Bases in RNA Read 1": "Q30 read1",
            "Q30 Bases in RNA Read 2": "Q30 read2",
        }
        data, self.vdj_mapping_headers = update_dict(
            data_general_stats,
            self.vdj_mapping_headers,
            data_rows,
            col_dict,
            {},
            "VDJ",
        )

        # Store VDJ annotation and expression data from vdj report
        col_dict = {
            "Cells With Productive V-J Spanning Pair": "cells VJ span",
            "Cells With Productive V-J Spanning (IGK, IGH) Pair": "cells IGK-IGH span",
            "Cells With Productive V-J Spanning (IGL, IGH) Pair": "cells IGL-IGH span",
            "Paired Clonotype Diversity": "clonotype diversity",
            "Cells With IGH Contig": "cells IGH contig",
            "Cells With IGK Contig": "cells IGK contig",
            "Cells With IGL Contig": "cells IGL contig",
            "Cells With CDR3-annotated IGH Contig": "cells IGH CDR3",
            "Cells With CDR3-annotated IGK Contig": "cells IGK CDR3",
            "Cells With CDR3-annotated IGL Contig": "cells IGL CDR3",
            "Cells With V-J Spanning IGH Contig": "cells IGH VJ span",
            "Cells With V-J Spanning IGK Contig": "cells IGK VJ span",
            "Cells With V-J Spanning IGL Contig": "cells IGL VJ span",
            "Cells With Productive IGH Contig": "cells IGH productive",
            "Cells With Productive IGK Contig": "cells IGK productive",
            "Cells With Productive IGL Contig": "cells IGL productive",
        }
        colours = {
            # "cells VJ span": "",
            # "cells IGK-IGH span": "",
            # "cells IGL-IGH span": "",
            # "clonotype diversity": "",
            # "cells IGH contig": "",
            # "cells IGK contig": "",
            # "cells IGL contig": "",
            # "cells IGH CDR3": "",
            # "cells IGK CDR3": "",
            # "cells IGL CDR3": "",
            # "cells IGH VJ span": "",
            # "cells IGK VJ span": "",
            # "cells IGL VJ span": "",
            # "cells IGH productive": "",
            # "cells IGK productive": "",
            # "cells IGL productive": "",
        }
        data_annotations, self.vdj_annotations_headers = update_dict(
            data_general_stats,
            self.vdj_annotations_headers,
            mydict["summary_tab"]["vdj_annotation"]["table"]["rows"],
            col_dict,
            colours,
            "VDJ",
        )

        # Extract warnings if any
        warnings = dict()
        alarms_list = mydict["alarms"].get("alarms", [])
        for alarm in alarms_list:
            warnings[alarm["id"]] = "FAIL"
            self.vdj_warnings_headers[alarm["id"]] = {
                "title": clean_title_case(alarm["id"].replace("_", " ")),
                "description": alarm["title"],
                "bgcols": {"FAIL": "#f06807"},
            }

        # Extract data for plots
        help_dict = {x[0]: x[1][0] for x in mydict["summary_tab"]["cells"]["help"]["data"]}
        plots = {
            "bc": {
                "config": {
                    "id": "mqc_cellranger_vdj_bc_knee",
                    "title": f"Cell Ranger VDJ: {mydict['summary_tab']['cells']['barcode_knee_plot']['layout']['title']}",
                    "xlab": mydict["summary_tab"]["cells"]["barcode_knee_plot"]["layout"]["xaxis"]["title"],
                    "ylab": mydict["summary_tab"]["cells"]["barcode_knee_plot"]["layout"]["yaxis"]["title"],
                    "yLog": True,
                    "xLog": True,
                },
                "description": "Barcode knee plot",
                "helptext": help_dict["Barcode Rank Plot"],
            }
        }
        plots_data = {"bc": parse_bcknee_data(mydict["summary_tab"]["cells"]["barcode_knee_plot"]["data"], s_name)}

        if len(data) > 0:
            if s_name in self.cellrangervdj_general_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            self.add_data_source(f, s_name, module="cellranger", section="count")
            self.cellrangervdj_mapping[s_name] = data
            self.cellrangervdj_general_data[s_name] = data_general_stats
            self.cellrangervdj_annotations[s_name] = data_annotations
            if len(warnings) > 0:
                self.cellrangervdj_warnings[s_name] = warnings
            self.cellrangervdj_plots_conf = plots
            for k in plots_data.keys():
                if k not in self.cellrangervdj_plots_data.keys():
                    self.cellrangervdj_plots_data[k] = dict()
                self.cellrangervdj_plots_data[k].update(plots_data[k])
